/*
  CRF++ -- Yet Another CRF toolkit

  $Id: encoder.cpp,v 1.20 2006/08/07 02:35:35 taku Exp $;

  Copyright(C) 2005 Taku Kudo <taku@chasen.org>

  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or(at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "param.h"
#include "encoder.h"
#include "timer.h"
#include "tagger.h"
#include "lbfgs.h"
#include "common.h"
#include "feature_index.h"
#include "scoped_ptr.h"
#include "thread.h"
#include <fstream>

namespace CRFPP {

#ifdef CRFPP_USE_THREAD
  class encoder_thread_t: public thread {
  public:
    unsigned short start_i;
    unsigned short thread_num;
    size_t size;
    size_t err;
    size_t all;
    size_t zeroone;
    double f;
    TaggerImpl **x;
    std::vector<double> expected;
    void run() {
      err = all = zeroone = 0;
      f = 0.0;
      std::fill(expected.begin(), expected.end(), 0.0);
      for (size_t i = start_i; i < size; i += thread_num) {
        f += x[i]->gradient(&(expected[0]));
        x[i]->eval(err, all, zeroone);
      }
    }
  };
#endif

  bool Encoder::convert(const char* textfilename,
                        const char *binaryfilename)
  {
    EncoderFeatureIndex feature_index(1);
    CHECK_FALSE(feature_index.convert(textfilename, binaryfilename))
      << feature_index.what();

    return true;
  }

  bool Encoder::learn(const char *templfile,
                      const char *trainfile,
                      const char *modelfile,
                      bool textmodelfile,
                      size_t maxitr,
                      size_t freq,
                      double eta,
                      double C,
                      unsigned short thread_num)
  {
    std::cout << COPYRIGHT << std::endl;

    EncoderFeatureIndex feature_index(thread_num);
    std::vector<TaggerImpl* > x;

    std::cout.setf(std::ios::fixed, std::ios::floatfield);
    std::cout.precision(5);

#define WHAT_ERROR(msg) { \
  WHAT << msg; \
  for (std::vector<TaggerImpl *>::iterator it = x.begin(); it != x.end(); ++it)  \
    delete *it; \
  x.clear(); \
  return false; }

    CHECK_FALSE(feature_index.open(templfile, trainfile)) << feature_index.what();

    {
      progress_timer pg;

      std::ifstream ifs(trainfile);
      CHECK_FALSE(ifs) << "cannot open: " << trainfile;

      std::cout << "reading training data: " << std::flush;
      size_t line = 0;
      while (ifs) {
        TaggerImpl *_x = new TaggerImpl();
        _x->open(&feature_index);
        _x->set_thread_id(line % thread_num);
        if (! _x->read (ifs) || ! _x->shrink())
          WHAT_ERROR(_x->what());

        if (! _x->empty()) x.push_back (_x);
        else delete _x;

        if (++line % 100 == 0) std::cout << line << ".. " << std::flush;
      }

      ifs.close();
      std::cout << "\nDone! ";
    }

    feature_index.shrink(freq);

    progress_timer pg;
    std::vector <double> alpha(feature_index.size());           // parameter
    std::vector <double> expected(feature_index.size());           // gradient
    std::fill(alpha.begin(), alpha.end(), 0.0);

    feature_index.set_alpha(&alpha[0]);

    double old_f = 1e+37;
    int    converge = 0;
    LBFGS lbfgs;
    lbfgs.init(feature_index.size(), 5);

    std::cout << "Number of sentences:          " << x.size() << std::endl;
    std::cout << "Number of features:           " << feature_index.size() << std::endl;
    std::cout << "Number of thread(s):          " << thread_num << std::endl;
    std::cout << "Freq:                         " << freq << std::endl;
    std::cout << "eta:                          " << eta << std::endl;
    std::cout << "C(sigma^2):                   " << C << std::endl << std::endl;

#ifdef CRFPP_USE_THREAD

    scoped_array<encoder_thread_t> thread(new encoder_thread_t [thread_num]);
    for (size_t i = 0; i < thread_num; i++) {
      thread[i].start_i = i;
      thread[i].size = x.size();
      thread[i].thread_num = thread_num;
      thread[i].x = &x[0];
      thread[i].expected.resize(expected.size());
    }
#endif

    for (size_t itr = 0; itr < maxitr; ++itr) {

      double f = 0.0;
      size_t err = 0;
      size_t all = 0;
      size_t zeroone = 0;

      std::fill(expected.begin(), expected.end(), 0.0);
      feature_index.clear();

#ifdef CRFPP_USE_THREAD
      for (size_t i = 0; i < thread_num; ++i)
        thread[i].start();

      for (size_t i = 0; i < thread_num; ++i)
        thread[i].join();

      for (size_t i = 0; i < thread_num; ++i) {
        f += thread[i].f;
        err += thread[i].err;
        all += thread[i].all;
        zeroone += thread[i].zeroone;
        for (size_t k = 0; k < feature_index.size(); ++k)
          expected[k] += thread[i].expected[k];
      }
#else
      for (size_t i = 0; i < x.size(); ++i) {
        f += x[i]->gradient(&expected[0]);
        x[i]->eval(err, all, zeroone);
      }
#endif

      for (size_t k = 0; k < feature_index.size(); ++k) {
        f +=(alpha[k] * alpha[k] /(2.0 * C));
        expected[k] += alpha[k]/C;
      }

      double diff =(itr == 0 ? 1.0 : std::abs(old_f - f)/old_f);
      std::cout << "iter="  << itr
                << " terr=" << 1.0 * err/all
                << " serr=" << 1.0 * zeroone/x.size()
                << " obj=" << f
                << " diff="  << diff << std::endl;
      old_f = f;

      if (diff < eta) converge++; else converge = 0;
      if (itr > maxitr || converge == 3)  break; // 3 is ad-hoc

      int ret = lbfgs.optimize(&alpha[0], &f, &expected[0]);
      if (ret <= 0) WHAT_ERROR(lbfgs.what());
    }

    if (! feature_index.save (modelfile, textmodelfile))
      WHAT_ERROR(feature_index.what());

    std::cout << "\nDone! ";

    return true;
#undef WHAT_ERROR
  }
}

int crfpp_learn(int argc, char **argv)
{
  static const CRFPP::Option long_options[] =
    {
      {"freq",     'f', "1",      "INT",   "use features that occuer no less than INT(default 1)" },
      {"maxiter" , 'm', "100000", "INT",   "set INT for max iterations in LBFGS routine(default 10k)" },
      {"cost",     'c', "1.0",    "FLOAT", "set FLOAT for cost parameter(default 1.0)" },
      {"eta",      'e', "0.0001", "FLOAT", "set FLOAT for termination criterion(default 0.0001)" },
      {"convert",  'C',  0,       0,        "convert text model to binary model" },
      {"textmodel", 't', 0,       0,       "build also text model file for debugging" },
#ifdef CRFPP_USE_THREAD
      {"thread", 'p',   "1",       "INT",   "number of threads(default 1)" },
#endif
      {"version",  'v', 0,        0,       "show the version and exit" },
      {"help",     'h', 0,        0,       "show this help and exit" },
      {0,0,0,0,0}
    };

  CRFPP::Param param;

  param.open(argc, argv, long_options);

  const std::vector<std::string> &rest = param.rest_args();

  if (param.getProfileInt("version")) {
    std::cout << param.version(long_options);
    return 0;
  }

  bool   convert = param.getProfileFloat("convert");

  if (param.getProfileInt("help") ||
      (convert && rest.size() != 2) ||(!convert && rest.size() != 3)) {
    std::cout << param.help(long_options);
    return 0;
  }

  size_t freq    = param.getProfileInt("freq");
  size_t maxiter = param.getProfileInt("maxiter");
  double C       = param.getProfileFloat("cost");
  double eta     = param.getProfileFloat("eta");
  bool   textmodel = param.getProfileFloat("textmodel");
  unsigned short thread = 1;
#ifdef CRFPP_USE_THREAD
  thread = param.getProfileInt("thread");
#endif

  CRFPP::Encoder encoder;
  if (convert) {
    if (! encoder.convert(rest[0].c_str(), rest[1].c_str())) {
      std::cerr << encoder.what() << std::endl;
      return -1;
    }
  } else {
    if (! encoder.learn (rest[0].c_str(),
                         rest[1].c_str(),
                         rest[2].c_str(),
                         textmodel,
                         maxiter, freq, eta, C, thread)) {
      std::cerr << encoder.what() << std::endl;
      return -1;
    }
  }

  return 0;
}
